"""
conversion
=========================

This module loads the survey data from 2022 from the survey database
and stores the data in the json structure of the new survey, so that
it can be used to prefill the 2023 survey.

"""
import logging
import click
import json

from sqlalchemy import delete, text, select

import compendium_v2
from compendium_v2.environment import setup_logging
from compendium_v2.db import db
from compendium_v2.config import load
from compendium_v2.survey_db import model as survey_model
from compendium_v2.db.model import NREN
from compendium_v2.db.survey_model import Survey, SurveyResponse, ResponseStatus

from compendium_v2.conversion import mapping

setup_logging()

logger = logging.getLogger('conversion')


def query_nren(nren_id: int):
    query = mapping.ANSWERS_2022_QUERY.format(nren_id)
    answers = {}
    for row in db.session.execute(text(query), bind_arguments={'bind': db.engines[survey_model.SURVEY_DB_BIND]}):
        answers[row[0]] = row[1]
    return answers


def convert_answers(answers):
    data = {}

    for id, question_name in mapping.ID_TO_NAME.items():
        if id not in answers:
            continue

        answer = answers[id]
        if len(answer) > 1 and answer[0] == '"' and answer[-1] == '"':
            answer = answer[1:-1]
            if id in mapping.VALUE_TO_CODE_MAPPING:
                answer = mapping.VALUE_TO_CODE_MAPPING[id][answer]
        if len(answer) > 1 and answer[0] == '[' and answer[-1] == ']':
            answer = json.loads(answer)
            if id in mapping.VALUE_TO_CODE_MAPPING:
                mapped_answer = []
                for entry in answer:
                    mapped_answer.append(mapping.VALUE_TO_CODE_MAPPING[id][entry])
                answer = mapped_answer

        if id == 16504:
            answer = '\n'.join(answer)

        # code to convert my description in the mapping to a json structure
        question_names = question_name.split(":")
        subdict = data
        for name in question_names[0:-1]:
            if name[-1] == "]":
                index = name[-2]
                if index == "[":
                    subdict = subdict.setdefault(name[:-2], [])
                    break  # special case where json list is mapped to a list of dicts (part 1)
                sublist = subdict.setdefault(name[:-3], [])
                index = int(index)
                while len(sublist) <= index:
                    sublist.append({})
                subdict = sublist[index]
            else:
                subdict = subdict.setdefault(name, {})

        if isinstance(subdict, list):  # special case where json list is mapped to a list of dicts (part 2)
            for answer_entry in answer:
                subdict.append({question_names[-1]: answer_entry})
        elif question_names[-1] == "available":  # special case because we changed the policies questions a bit
            if answer == "Yes":
                subdict[question_names[-1]] = ["yes"]
        else:
            subdict[question_names[-1]] = answer

    for id, question_name in mapping.ID_TO_NAME_SERVICES.items():
        if id not in answers:
            continue

        answer = answers[id]
        answer = json.loads(answer)
        for user_type in answer:
            user_type_code = mapping.SERVICE_USER_TYPE_TO_CODE[user_type]
            formatted_question_name = question_name.format(user_type_code)

            question_names = formatted_question_name.split(":")
            subdict = data
            for name in question_names[0:-2]:
                subdict = subdict.setdefault(name, {})

            sublist = subdict.setdefault(question_names[-2], [])
            sublist.append(question_names[-1])

    return {"data": data}


def _cli(app):
    with app.app_context():

        nren_surveys = {}

        for nren in db.session.scalars(select(NREN)):
            survey_db_nren_id = mapping.NREN_IDS[nren.name]
            nren_surveys[nren] = query_nren(survey_db_nren_id)

        db.session.execute(delete(SurveyResponse).where(
            SurveyResponse.survey_year == 2022
        ))

        survey = db.session.scalar(select(Survey).where(Survey.year == 2022))

        for nren, answers in nren_surveys.items():
            survey_dict = convert_answers(answers)
            survey_dict["page"] = 0
            survey_dict["verification_status"] = {}
            response = SurveyResponse(
                nren=nren,
                nren_id=nren.id,
                survey_year=2022,
                survey=survey,
                answers=survey_dict,
                status=ResponseStatus.checked
            )
            db.session.add(response)

        db.session.commit()


@click.command()
@click.option('--config', type=click.STRING, default='config.json')
def cli(config):
    app_config = load(open(config, 'r'))

    app_config['SQLALCHEMY_BINDS'] = {survey_model.SURVEY_DB_BIND: app_config['SURVEY_DATABASE_URI']}

    app = compendium_v2._create_app_with_db(app_config)
    _cli(app)


if __name__ == "__main__":
    cli()
