import logging
from typing import Any

from flask import Blueprint, jsonify
from sqlalchemy import select

from compendium_v2.db import db
from compendium_v2.db.model import ChargingStructure
from compendium_v2.routes import common


routes = Blueprint('charging', __name__)
logger = logging.getLogger(__name__)

CHARGING_STRUCTURE_RESPONSE_SCHEMA = {
    '$schema': 'http://json-schema.org/draft-07/schema#',

    'definitions': {
        'charging': {
            'type': 'object',
            'properties': {
                'nren': {'type': 'string'},
                'nren_country': {'type': 'string'},
                'year': {'type': 'integer'},
                'fee_type': {'type': ["string", "null"]},
            },
            'required': ['nren', 'nren_country', 'year', 'fee_type'],
            'additionalProperties': False
        }
    },

    'type': 'array',
    'items': {'$ref': '#/definitions/charging'}
}


@routes.route('/', methods=['GET'])
@common.require_accepts_json
def charging_structure_view() -> Any:
    """
    handler for /api/charging/ requests

    response will be formatted as:

    .. asjson::
        compendium_v2.routes.charging.CHARGING_STRUCTURE_RESPONSE_SCHEMA

    :return:
    """

    def _extract_data(entry: ChargingStructure):
        return {
            'nren': entry.nren.name,
            'nren_country': entry.nren.country,
            'year': int(entry.year),
            'fee_type': entry.fee_type.value if entry.fee_type is not None else None,
        }

    entries = sorted(
        [_extract_data(entry) for entry in db.session.scalars(select(ChargingStructure))],
        key=lambda d: (d['nren'], d['year'])
    )
    return jsonify(entries)
