import logging

from flask import Blueprint, jsonify
from sqlalchemy import select

from compendium_v2.routes import common
from compendium_v2.db import db
from compendium_v2.db.model import FundingSource
from typing import Any


routes = Blueprint('funding', __name__)
logger = logging.getLogger(__name__)

FUNDING_RESPONSE_SCHEMA = {
    '$schema': 'http://json-schema.org/draft-07/schema#',

    'definitions': {
        'funding': {
            'type': 'object',
            'properties': {
                'nren': {'type': 'string'},
                'nren_country': {'type': 'string'},
                'year': {'type': 'integer'},
                'client_institutions': {'type': 'number'},
                'european_funding': {'type': 'number'},
                'gov_public_bodies': {'type': 'number'},
                'commercial': {'type': 'number'},
                'other': {'type': 'number'}
            },
            'required': ['nren', 'nren_country', 'year', 'client_institutions',
                         'european_funding', 'gov_public_bodies', 'commercial', 'other'],
            'additionalProperties': False
        }
    },

    'type': 'array',
    'items': {'$ref': '#/definitions/funding'}
}


@routes.route('/', methods=['GET'])
@common.require_accepts_json
def funding_source_view() -> Any:
    """
    handler for /api/funding/ requests

    response will be formatted as:

    .. asjson::
        compendium_v2.routes.funding.FUNDING_RESPONSE_SCHEMA

    :return:
    """

    def _extract_data(entry: FundingSource):
        return {
            'nren': entry.nren.name,
            'nren_country': entry.nren.country,
            'year': entry.year,
            'client_institutions': float(entry.client_institutions),
            'european_funding': float(entry.european_funding),
            'gov_public_bodies': float(entry.gov_public_bodies),
            'commercial': float(entry.commercial),
            'other': float(entry.other)
        }

    entries = sorted(
        [_extract_data(entry) for entry in db.session.scalars(select(FundingSource))],
        key=lambda d: (d['nren'], d['year'])
    )
    return jsonify(entries)
