import logging
from typing import Any

from flask import Blueprint, jsonify
from sqlalchemy import select
from compendium_v2.db import db
from compendium_v2.db.model import Policy, NREN
from compendium_v2.routes import common

routes = Blueprint('policy', __name__)
logger = logging.getLogger(__name__)

POLICY_RESPONSE_SCHEMA = {
    '$schema': 'http://json-schema.org/draft-07/schema#',
    'definitions': {
        'policy': {
            'type': 'object',
            'properties': {
                'nren': {'type': 'string'},
                'nren_country': {'type': 'string'},
                'year': {'type': 'integer'},
                'strategic_plan': {'type': 'string'},
                'environmental': {'type': 'string'},
                'equal_opportunity': {'type': 'string'},
                'connectivity': {'type': 'string'},
                'acceptable_use': {'type': 'string'},
                'privacy_notice': {'type': 'string'},
                'data_protection': {'type': 'string'},
            },
            'required': ['nren', 'nren_country', 'year', 'strategic_plan', 'environmental', 'equal_opportunity',
                         'connectivity', 'acceptable_use', 'privacy_notice', 'data_protection'],
            'additionalProperties': False
        }
    },

    'type': 'array',
    'items': {'$ref': '#/definitions/policy'}
}


@routes.route('/', methods=['GET'])
@common.require_accepts_json
def policy_view() -> Any:
    """
    handler for /api/policy/ requests
    returns policy information for each NREN/year combination

    response will be formatted as:

    .. asjson::
        compendium_v2.routes.policy.POLICY_RESPONSE_SCHEMA

    :return:
    """

    def _extract_data(entry: Policy):
        return {
            'nren': entry.nren.name,
            'nren_country': entry.nren.country,
            'year': entry.year,
            'strategic_plan': entry.strategic_plan,
            'environmental': entry.environmental,
            'equal_opportunity': entry.equal_opportunity,
            'connectivity': entry.connectivity,
            'acceptable_use': entry.acceptable_use,
            'privacy_notice': entry.privacy_notice,
            'data_protection': entry.data_protection,
        }

    entries = [_extract_data(entry) for entry in db.session.scalars(
        select(Policy).join(NREN).order_by(NREN.name.asc(), Policy.year.desc()))]
    return jsonify(entries)
