import React from "react";
import { Question, FunctionFactory } from "survey-core";
import { Survey } from "survey-react-ui";
import { VerificationStatus } from './Schema';

function SurveyComponent({ surveyModel, verificationStatus }) {

    function validateWebsiteUrl(params) {
        const value = params[0];
        if (value === undefined || value == null || value == '') {
            return true;
        }
        try {
            const url = new URL(value);
            return url.protocol === 'http:' || url.protocol === 'https:';
        } catch (err) {
            return false;
        }
    }

    surveyModel.css = {
        question: {
            title: "sv-title sv-question__title sv-header-flex",
            titleOnError: "sv-question__title--error sv-error-color-fix"
        }
    };

    function setVerifyButton(question: Question, state: VerificationStatus) {

        verificationStatus.current.set(question.name, state);

        const btn = document.createElement("button");
        btn.type = "button";
        btn.className = "sv-action-bar-item verification";
        btn.innerHTML = state;

        if (state == VerificationStatus.Unverified) {
            btn.innerHTML = "No change from previous year";
            btn.className += " verification-required";
            btn.onclick = function () {
                question.validate();
                setVerifyButton(question, VerificationStatus.Verified);
            }
        } else {
            btn.innerHTML = "Answer updated"
            btn.className += " verification-ok";
        }

        const selector = '[data-name="' + question.name + '"]';
        const header = document.querySelector(selector)?.querySelector('h5');

        const old = header?.querySelector(".verification");
        if (old) {
            old.replaceWith(btn);
        } else {
            header?.appendChild(btn);
        }
    }

    FunctionFactory.Instance.register("validateWebsiteUrl", validateWebsiteUrl);

    surveyModel.onAfterRenderQuestion.add(function (survey, options) {
        const status = verificationStatus.current.get(options.question.name);
        if (status) {
            setVerifyButton(options.question, status);
        }
    });

    surveyModel.onValueChanged.add(function (survey, options) {
        const currentStatus = verificationStatus.current.get(options.question.name);
        if (currentStatus == VerificationStatus.New) {
            setVerifyButton(options.question, VerificationStatus.Answered);
        } else if (currentStatus == VerificationStatus.Unverified) {
            setVerifyButton(options.question, VerificationStatus.Edited);
        }
    });

    surveyModel.onUpdateQuestionCssClasses.add(function (_, options) {
        if (options.question.hideCheckboxLabels) {
            const classes = options.cssClasses;
            classes.root += " hidden-checkbox-labels";
        }
    });

    surveyModel.onMatrixAfterCellRender.add((survey, options) => {
        // get the customDescription for matrix rows and set it in the title
        // attribute so that it shows up as a hover popup
        // NB I would have preferred using onAfterRenderQuestion, but unfortunately that is
        // not always triggered on re-renders (specifically when extra column become visble or invisible)

        if (options.column['indexValue'] == 0 && 'item' in options.row) {
            const item = options.row['item'] as object;
            if (item['customDescription'] !== undefined) {
                options.htmlElement.parentElement?.children[0].setAttribute("title", item['customDescription']);
            }
        }
    });

    return <Survey model={surveyModel} />
}

export default SurveyComponent;
