import React, { useEffect, useRef, useState } from "react";
import toast, { Toaster } from "react-hot-toast";
import { Model, Serializer } from "survey-core";
import { useParams } from "react-router-dom";
import SurveyComponent from "./SurveyComponent";
import SurveyNavigationComponent from "./SurveyNavigationComponent";
import { VerificationStatus } from './Schema';
import "survey-core/modern.min.css";
import './survey.scss';
import { Container, Row } from "react-bootstrap";

Serializer.addProperty("itemvalue", "customDescription:text");
Serializer.addProperty("question", "hideCheckboxLabels:boolean");


function SurveyContainerComponent({ loadFrom, saveTo = '', readonly = false }) {
    const [surveyModel, setSurveyModel] = useState<Model>();
    const verificationStatus = useRef<Map<string, VerificationStatus>>(new Map());
    const { year, nren } = useParams();
    const [error, setError] = useState<string>('loading survey...');

    useEffect(() => {
        async function getModel() {
            const response = await fetch(loadFrom + year + (nren ? '/' + nren : ''))  // year is always set, nren stays empty for inspect and try
            const json = await response.json();

            if (!response.ok) {

                if ('message' in json) {
                    throw new Error(json.message);
                } else {
                    throw new Error(`Request failed with status ${response.status}`);
                }
            }


            for (const questionName in json["verification_status"]) {
                verificationStatus.current.set(questionName, json["verification_status"][questionName]);
            }

            const survey = new Model(json['model']);
            survey.setVariable('surveyyear', year);
            survey.setVariable('previousyear', parseInt(year!) - 1);

            survey.data = json['data'];
            survey.clearIncorrectValues(true);  // TODO test if this really removes all old values and such
            survey.currentPageNo = json['page'];
            survey.showNavigationButtons = false;
            survey.showTOC = false;
            if (readonly) {
                survey.mode = 'display';
            }

            setSurveyModel(survey);
        }

        getModel().catch(error => setError('Error when loading survey: ' + error.message))

    }, []);

    if (!surveyModel) {
        return error
    }

    function saveSurveyData(survey, success?, failure?) {
        if (saveTo == '') {
            return;
        }
        const xhr = new XMLHttpRequest();
        xhr.open("POST", saveTo + year + '/' + nren);
        xhr.setRequestHeader("Content-Type", "application/json; charset=utf-8");
        xhr.onload = xhr.onerror = () => {
            if (xhr.status == 200 && success) {
                success();
            } else if (xhr.status != 200 && failure) {
                failure();
            }
        }
        const saveData = {
            data: survey.data,
            page: survey.currentPageNo,
            verification_status: Object.fromEntries(verificationStatus.current)
        }
        xhr.send(JSON.stringify(saveData));
    }

    const saveSurvey = () => {
        saveSurveyData(surveyModel, () => toast("Survey saved!"), () => toast("Failed saving survey!")); 
    };

    const endSurvey = () => {
        let firstValidationError = '';
        const verificationValidator = (survey, options) => {
            const status = verificationStatus.current.get(options.name);
            if (status == VerificationStatus.Unverified) {
                if (firstValidationError == '') {
                    firstValidationError = options.name;
                }
                options.error = 'Please verify that last years data is correct by editing the answer or pressing the "No change from previous year" button!';
            }
        };
        surveyModel.onValidateQuestion.add(verificationValidator);
        const validSurvey = surveyModel.validate();
        surveyModel.onValidateQuestion.remove(verificationValidator);
        if (validSurvey) {
            saveSurveyData(surveyModel, () => toast("Survey completed!"), () => toast("Failed completing survey!"));
        } else {
            surveyModel.focusQuestion(firstValidationError);
        }
    };

    const validatePage = () => {
        // TODO remove duplication with funtion above and perhaps also first focus normal validation errors and only afterwards the verification validation errors
        let firstValidationError = '';
        const verificationValidator = (survey, options) => {
            const status = verificationStatus.current.get(options.name);
            if (status == VerificationStatus.Unverified) {
                if (firstValidationError == '') {
                    firstValidationError = options.name;
                }
                options.error = 'Please verify that last years data is correct by editing the value or pressing the verification button!';
            }
        };
        surveyModel.onValidateQuestion.add(verificationValidator);
        const validSurvey = surveyModel.validateCurrentPage();
        surveyModel.onValidateQuestion.remove(verificationValidator);
        if (validSurvey) {
            // TODO some notification
        } else {
            surveyModel.focusQuestion(firstValidationError);
        }
    }

    return (
        <Container className="survey-container">
            <Toaster />
            <Row className="survey-content">
                <h2><span className="survey-title">{year} Compendium Survey</span><span className="survey-title-nren">{nren}</span></h2>
                <span className="survey-description">How to complete this survey -
                    Lorem ipsum dolor sit amet, consectetur adipiscing elit. Etiam commodo tempus pulvinar. Aenean interdum lectus nec diam porttitor interdum.
                    Sed efficitur, magna ut dignissim consequat, purus lacus aliquam tortor, ut dignissim orci ligula ac ligula.</span>

            </Row>
            <Row>
                <SurveyNavigationComponent surveyModel={surveyModel} endSurvey={endSurvey} saveSurvey={saveSurvey} validatePage={validatePage} readonly={readonly}>
                    <SurveyComponent surveyModel={surveyModel} verificationStatus={verificationStatus} />
                </SurveyNavigationComponent>
            </Row>
        </Container>
    );
}

export default SurveyContainerComponent;
