import json
import jsonschema
from compendium_v2.db import db
from compendium_v2.db.survey_model import Survey, SurveyStatus
from compendium_v2.routes.survey import LIST_SURVEYS_RESPONSE_SCHEMA, SURVEY_RESPONSE_SCHEMA, VerificationStatus


def test_survey_route_list_response(client, test_survey_data):
    rv = client.get(
        '/api/survey/list',
        headers={'Accept': ['application/json']})
    assert rv.status_code == 200
    result = json.loads(rv.data.decode('utf-8'))
    jsonschema.validate(result, LIST_SURVEYS_RESPONSE_SCHEMA)
    assert result


def test_survey_route_new(app, client, test_survey_data, mocked_user):
    rv = client.post(
        '/api/survey/new',
        headers={'Accept': ['application/json']})
    assert rv.status_code == 400
    result = json.loads(rv.data.decode('utf-8'))
    assert not result.get('success')

    # mark all surveys as published
    with app.app_context():
        for survey in db.session.query(Survey).all():
            survey.status = SurveyStatus.published
        db.session.commit()

    rv = client.post(
        '/api/survey/new',
        headers={'Accept': ['application/json']})
    assert rv.status_code == 200
    result = json.loads(rv.data.decode('utf-8'))
    assert result == {'success': True}

    rv = client.post(
        '/api/survey/new',
        headers={'Accept': ['application/json']})
    assert rv.status_code != 200


def test_survey_route_open_close(app, client, test_survey_data, mocked_user):
    rv = client.post(
        '/api/survey/new',
        headers={'Accept': ['application/json']})
    assert rv.status_code == 400
    result = json.loads(rv.data.decode('utf-8'))
    assert not result.get('success')

    # mark all surveys as published
    with app.app_context():
        for survey in db.session.query(Survey).all():
            survey.status = SurveyStatus.published
        db.session.commit()

    rv = client.post(
        '/api/survey/new',
        headers={'Accept': ['application/json']})
    assert rv.status_code == 200
    result = json.loads(rv.data.decode('utf-8'))
    assert result.get('success')

    rv = client.post(
        '/api/survey/open/2023',
        headers={'Accept': ['application/json']})
    assert rv.status_code == 200
    result = json.loads(rv.data.decode('utf-8'))
    assert result == {'success': True}

    rv = client.post(
        '/api/survey/open/2023',
        headers={'Accept': ['application/json']})
    assert rv.status_code != 200

    rv = client.post(
        '/api/survey/close/2023',
        headers={'Accept': ['application/json']})
    assert rv.status_code == 200
    result = json.loads(rv.data.decode('utf-8'))
    assert result == {'success': True}

    rv = client.post(
        '/api/survey/close/2023',
        headers={'Accept': ['application/json']})
    assert rv.status_code != 200


def test_survey_route_publish(app, client, test_survey_data, mocked_admin_user):
    rv = client.post(
        '/api/survey/publish/2022',
        headers={'Accept': ['application/json']})
    assert rv.status_code == 400
    result = json.loads(rv.data.decode('utf-8'))
    assert not result.get('success')

    with app.app_context():
        survey = db.session.query(Survey).filter(Survey.year == 2022).first()
        survey.status = SurveyStatus.closed
        db.session.commit()

    rv = client.post(
        '/api/survey/publish/2022',
        headers={'Accept': ['application/json']})
    assert rv.status_code == 200
    result = json.loads(rv.data.decode('utf-8'))
    assert result.get('success')


def test_survey_route_try_response(client, test_survey_data):
    rv = client.get(
        '/api/survey/try/2022',
        headers={'Accept': ['application/json']})
    assert rv.status_code == 200
    result = json.loads(rv.data.decode('utf-8'))
    jsonschema.validate(result, SURVEY_RESPONSE_SCHEMA)
    assert result


def test_survey_route_inspect_response(client, test_survey_data):
    rv = client.get(
        '/api/survey/inspect/2022',
        headers={'Accept': ['application/json']})
    assert rv.status_code == 200
    result = json.loads(rv.data.decode('utf-8'))
    jsonschema.validate(result, SURVEY_RESPONSE_SCHEMA)
    assert result


def test_survey_route_save_load_response(client, test_survey_data, mocked_user):
    rv = client.post(
        '/api/survey/save/2021/nren2',
        headers={'Accept': ['application/json']},
        json={
            'data': {'q1': 'yes', 'q2': ['no']},
            'page': 3,
            'verification_status': {'q1': VerificationStatus.Verified}
        })
    assert rv.status_code == 400
    result = json.loads(rv.data.decode('utf-8'))
    assert not result.get('success')
    assert result.get('message') == 'Survey is closed'

    rv = client.get(
        '/api/survey/load/2021/nren2',
        headers={'Accept': ['application/json']})
    assert rv.status_code == 200
    result = json.loads(rv.data.decode('utf-8'))
    jsonschema.validate(result, SURVEY_RESPONSE_SCHEMA)
    assert result['page'] == 0
    assert result['data'] == {}
    assert result['verification_status'] == {}

    rv = client.post(
        '/api/survey/save/2022/nren2',
        headers={'Accept': ['application/json']},
        json={
            'data': {'q1': 'yes', 'q2': ['no']},
            'page': 3,
            'verification_status': {'q1': VerificationStatus.Verified}
        })
    assert rv.status_code == 200
    result = json.loads(rv.data.decode('utf-8'))
    assert result.get('success')

    rv = client.get(
        '/api/survey/load/2022/nren2',
        headers={'Accept': ['application/json']})
    assert rv.status_code == 200
    result = json.loads(rv.data.decode('utf-8'))
    jsonschema.validate(result, SURVEY_RESPONSE_SCHEMA)
    assert result['page'] == 3
    assert result['data'] == {'q1': 'yes', 'q2': ['no']}
    assert result['verification_status'] == {'q1': VerificationStatus.Verified}
